/******************************************************************************
 * Box.cpp
 *
 * GDSAT
 *
 ******************************************************************************
 * Copyright (C) 2013 Benjamin Niedermann <niedermann@kit.edu>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program.  If not, see <http://www.gnu.org/licenses/>.
 *****************************************************************************/

#include "Box.h"

Box::Box() {}

Box::~Box() {}

void Box::init(CNFModel & model, int * size, int dim){


	this->dimCount = dim;
	this->size = size;
	length =1;
	b = new Range[dim];
	e = new Range[dim];
	for(int i=0; i < dim; ++i){
		length *= size[i];
		b[i] = model.reserveVariables(size[i]);
		e[i] = model.reserveVariables(size[i]);
	}
	X = model.reserveVariables(length);
	createConstraints(model);
}


void Box::createConstraints(CNFModel & model){

	// CONSTRAINT (4): enforce frame
	enforceFrame(model);

	// CONSTRAINT (6) and (7): uniqueness of b- and e-variables.
	enforceUniqueness(model);

	// CONSTRAINT (5): there are no empty boxes.
	Clause * cnf2 = model.next(length);
	for(int x=X.begin; x < X.end; ++x){
		cnf2->add(x,true);
	}

	// CONSTRAINT (8) and (9): relation b,e and x variables.
	int count =0;
	for(int i=0; i<  dimCount; ++i){
		int pos =0;
		for(int x=X.begin; x < X.end; ++x,++pos){
			if(hasPre(x,i)){
				Clause * cnf = model.next(3);
				cnf->add(x,false);
				cnf->add(pre(x,i),true);
				cnf->add(getB(pos,i),true);


			}
			if(hasSuc(x,i)){
				Clause * cnf = model.next(3);
				cnf->add(x,false);
				cnf->add(suc(x,i),true);
				cnf->add(getE(pos,i),true);
			}
			count++;
		}
	}

}



void Box::enforceFrame(CNFModel & model){
	if(dimCount == 1){
		Clause * cnf = model.next(1);
		cnf->add(getX(0),false);
		cnf = model.next(1);
		cnf->add(getX(length-1),false);
		return;
	}
	if(dimCount == 2){
		for(int row=0; row < size[1]; row++){
			if(row==0 || row  == size[1]-1){

				int x = get(row);

				for(int col=0; col < size[0]; ++col, ++x){
					Clause * cnf = model.next(1);
					cnf->add(x,false);

				}
			}else{
				Clause * cnf = model.next(1);
				cnf->add(get(row,0),false);
				cnf = model.next(1);
				cnf->add(get(row,size[0]-1),false);

			}

		}
		return;
	}
	throw "Dimension is not supported";

}

string Box::toString(CNFModel & model){
	if(dimCount == 1){
		throw "not yet implemented";
	}
	if(dimCount ==2){
		stringstream stream;
		stream << "begin: " << get() << " end: " << end()<<endl;
		for(int row=-1; row < size[1]+1; ++row){
			if(row == -1){
				stream << "_ ";
				for(int i=0; i < size[0]; ++i){
					stream << model.getValue(b[0].begin+i) << " ";
				}
				stream << "_";
			}else if(row < size[1]){
				for(int col=-1; col < size[0]+1; ++col){
					if(col == -1){
						stream << model.getValue(b[1].begin+row) << " ";
					}else if(col < size[0]){
						stream << model.getValue(get(row,col)) << " ";
					}else{
						stream << model.getValue(e[1].begin+row) << " ";
					}
				}
			}else{
				stream << "_ ";
				for(int i=0; i < size[0]; ++i){
					stream << model.getValue(e[0].begin+i) << " ";
				}
				stream << "_";
			}
			stream << endl;
		}
		return stream.str();
	}
	throw "not yet implemented";
}


void Box::enforceUniqueness(CNFModel & model){
	for(int i=0; i < dimCount; ++i){
		enforceUniqueness(model,b[i]);
		enforceUniqueness(model,e[i]);
	}
}

void Box::enforceUniqueness(CNFModel & model, Range & vars){

	for(int x=vars.begin; x < vars.end; ++x){
		for(int y = x+1; y < vars.end; ++y){
			Clause * cnf = model.next(2);
			cnf->add(x,false);
			cnf->add(y,false);
		}
	}
}
