/******************************************************************************
 * Box.h
 *
 * GDSAT
 *
 ******************************************************************************
 * Copyright (C) 2013 Benjamin Niedermann <niedermann@kit.edu>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program.  If not, see <http://www.gnu.org/licenses/>.
 *****************************************************************************/

#ifndef BOX_H_
#define BOX_H_

#include "Model.h"

/**
 * Represents a box (see Paper).
 */
class Box {


	int   * size;
	int   length;
	int   dimCount;
	Range X;
	Range *b;
	Range *e;

	/**
	 * Enforces that the box does not occupy grid cells on the frame
	 * of the grid.
	 */
	void enforceFrame(CNFModel & model);

	/** Enforces that for each dimension there is only variable b and one variable e
	 *  is set true.
	 */
	void enforceUniqueness(CNFModel & model);
	void enforceUniqueness(CNFModel & mode, Range & vars);

	/**
	 * Creates the constraints necessary for a box.
	 */
	void createConstraints(CNFModel & model);
public:
	Box();
	virtual ~Box();



	/**
	 * Estimates the number of clauses necessary for a box
	 * of given dimension and size.
	 * - dim: number of dimensions of the box.
	 * - sizes: array containing for each dimension the length of the box.
	 */
	static int cnfCount(int dim, int * sizes){


		int length =1;
		int frame =0;
		int uniqueness =0;
		for(int i=0; i < dim; i++){
			frame += sizes[i]*2;
			uniqueness += 2* sizes[i]*(sizes[i]-1)/2;
			length *= sizes[i];
		}


		return 1 + 2*length*dim+frame+uniqueness;
	}

	/**
	 * Estimates the number of variables necessary for a box
	 * of given dimension and size.
	 * - dim: number of dimensions of the box.
	 * - sizes: array containing for each dimension the length of the box.
	 */
	static int varCount(int dim,int * sizes){

		int BandE =0;
		int x =1;
		for(int i=0; i < dim; i++){
			x *= sizes[i];
			BandE += sizes[i]*2;
		}
		return BandE+x;
	}

	/**
	 * Initializes a box. In particular cnfs are created and stored in the
	 * given model representing the box.
	 * - model: the cnf model used for creating the box.
	 * - sizes: array containing for each dimension the length of the box.
	 * - dim: number of dimensions of the box.
	 */
	void init(CNFModel & model, int * size, int dim);

	/**
	 * Given an already solved model this method returns a string representation
	 * of this box regarding the given model.
	 * The method requires that the box has been initialized with the given model.
	 */
	string toString(CNFModel & model);



	/**
	 * Returns the column of the variable.
	 */
	int getCol(int var){
		if(dimCount == 1){
			return var-X.begin;
		}
		if(dimCount == 2){
			int offset = var-X.begin;
			return offset - (offset/size[0]);
		}
		throw "Dimension is not supported.";
	}

	/**
	 * Returns the row of the variable.
	 */
	int getRow(int var){
		if(dimCount == 1){
			return 0;
		}
		if(dimCount == 2){
			int offset = var-X.begin;
			return offset/size[0];
		}
		throw "Dimension is not supported.";
	}



	/**
	 * Checks for a given dimension whether the given variable has a predecessor.
	 * Example:
	 * 1.) For a two dimensional box the call hasPre(var,0) checks whether
	 * the variable is not the first variable in its row.
	 * 2.) For a two dimensional box the call hasPre(var,1) checks whether
	 * the variable is not located in the first row.
	 */
	bool hasPre(int var, int dim){
		if(dim == 0){return getCol(var) != 0;}
		if(dim == 1){return getRow(var) != 0;}
		throw "not supported dimension";
	}

	/**
	 * Checks for a given dimension whether the given variable has a successor.
	 * Example: For a two dimensional box the call hasSuc(var,0) checks whether
	 * the variable is not the last variable in its row.
	 */
	bool hasSuc(int var, int dim){
		if(dim == 0){return getCol(var) != size[0]-1;}
		if(dim == 1){return getRow(var) != size[0]-1;}
		throw "not supported dimension";
	}

	/**
	 * Returns the predecessor of a variable:
	 * dim=0: predecessor of var is the variable in the previous column
	 * but in the same row.
	 * dim=1: predecessor of var is the variable in the same column
	 * but in the previous row.
	 */
	int pre(int var, int dim){
		if(dim == 0){ return var-1;}
		if(dim == 1){ return var-size[0];}
		throw "not supported dimCount";
	}

	/**
	 * Returns the successor of a variable:
	 */
	int suc(int var, int dim){
		if(dim == 0){ return var+1;}
		if(dim == 1){ return var+size[0];}
		throw "not supported dimCount";
	}


	/**
	 * Returns the the b-variable for a certain position
	 * in a given dimension.
	 */
	int getB(int pos, int dim){
		if(dim == 0){return b[0].begin+pos%size[0];}
		if(dim == 1){return b[1].begin+pos/size[0];}
		throw "not supported dimCount";
	}

	/**
	 * Returns the the e-variable for a certain position
	 * in a given dimension.
	 */
	int getE(int pos, int dim){
		if(dim == 0){return e[0].begin+pos%size[0];}
		if(dim == 1){return e[1].begin+pos/size[0];}
		throw "not supported dimCount";
	}

	/**
	 * Methods for iterating through b and e variables.
	 */
	int beginB(int dim) {return b[dim].begin;}
	int endB(int dim) {return b[dim].end;}


	int beginE(int dim) {return e[dim].begin;}
	int endE(int dim) {return e[dim].end;}



	/**
	 * This method is especially useful for one-dimensional boxes
	 * in order to access the variables in X.
	 */
	int getX(int pos){
		return X.begin + pos;
	}

	/**
	 * Returns the whole range of variables in X.
	 */
	Range & getX(){
		return X;
	}

	int end() {return X.end;}
	int get(){return X.begin;}

	int get(int row, int col){
		return  X.begin+row * size[0] +col;
	}

	int get(int row){
		return X.begin+row*size[0];
	}
};

#endif /* BOX_H_ */
