/******************************************************************************
 * Pathwidth.cpp
 *
 * GDSAT
 *
 ******************************************************************************
 * Copyright (C) 2013 Benjamin Niedermann <niedermann@kit.edu>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program.  If not, see <http://www.gnu.org/licenses/>.
 *****************************************************************************/

#include <algorithm>
#include <iostream>


#include "PathWidth.h"


int factorialNew(int nValue)
   {
	   if(nValue == 1 || nValue == 2){return nValue;}
       int result = nValue;
       do
       {
    	   nValue--;
           result = result*nValue;

       }while(nValue>=2);
       return result;
   }

int evaluateBinomialCoefficientNew(int n, int k)
   {
	int r=1;

	if(n-k < k){
	   for (int i=1; i<= n-k; ++i ){
	      r = r*(1 + k/i);
	   }
	}else{
	   for (int i=1; i<= k; ++i){
	      r = r*(1 + (n-k)/i);
	   }

	}
	return r;
}

template <typename Iterator>
inline bool next_combinationNew(const Iterator first, Iterator k, const Iterator last)
{
   /* Credits: Thomas Draper */
   if ((first == last) || (first == k) || (last == k))
      return false;
   Iterator itr1 = first;
   Iterator itr2 = last;
   ++itr1;
   if (last == itr1)
      return false;
   itr1 = last;
   --itr1;
   itr1 = k;
   --itr2;
   while (first != itr1)
   {
      if (*--itr1 < *itr2)
      {
         Iterator j = k;
         while (!(*itr1 < *j)) ++j;
         std::iter_swap(itr1,j);
         ++itr1;
         ++j;
         itr2 = k;
         std::rotate(itr1,j,last);
         while (last != j)
         {
            ++j;
            ++itr2;
         }
         std::rotate(k,itr2,last);
         return true;
      }
   }
   std::rotate(first,k,last);
   return false;
}




PWModel * createPWModel(int pathwidth, Graph & graph, Clock& clock) {
	clock.start();

	int length = graph.n()+2;

	int * size = new int[1];
	size[0] = length;

	// upper bound for variables and cnfs.
	int varCount = Box::varCount(1,size)*graph.n() + graph.m()*length;
	int cnfCount = Box::cnfCount(1,size)*graph.n()
			        + graph.m()  // constraint 1
			        + graph.m()*length  // constraint 2
			        + graph.m()*length   // constraint 3
			        + evaluateBinomialCoefficientNew(graph.n(),pathwidth+1)*length*graph.n() // constraint 4
			        ;

	PWModel * pwModel = new PWModel(pathwidth,varCount,cnfCount);

	pwModel->length = length;

    CNFModel & model  = pwModel->getCNFModel();

    Box * nodes = new Box[graph.n()];
    for(int i=0; i < graph.n(); ++i){
    	nodes[i].init(model,size,1);
    }

    Range * edges = new Range[graph.m()];

    for(int i=0; i < graph.m(); ++i){
    	edges[i] = model.reserveVariables(length);
    }

    pwModel->edges = edges;
    pwModel->nodes = nodes;

    // CONSTRAINTS (11)-(12)
	for(int i=0; i < graph.m(); ++i){
		Clause* cnf = model.next(length);
		Edge edge = graph.edges()[i];

		for(int j=0; j < length; ++j){
			cnf->add(edges[i].getVar(j),true);   // constraint 1

			Clause * cnf1 = model.next(2);
			Clause * cnf2 = model.next(2);
			cnf1->add(edges[i].getVar(j),false); // constraint 2
			cnf1->add(nodes[edge.first].getX(j),true);

			cnf2->add(edges[i].getVar(j),false); // constraint 3
			cnf2->add(nodes[edge.second].getX(j),true);

		}
	}




	// CONSTRAINT (13)
	vector<int> nodeVec;
	nodeVec.resize(graph.n(),0);
	for(int i=0; i < graph.n(); ++i){
		nodeVec[i]=i;
	}



	int k = pathwidth+2;
	if(k <= graph.n()){
	do
	{
      for(int pos=0; pos < length; ++pos){
		  Clause * cnf = model.next(k);
		  for(int i=0; i < k; ++i){

			  cnf->add(nodes[nodeVec[i]].getX(pos),false);
		  }
       }
      }
	  while(next_combinationNew(nodeVec.begin(),nodeVec.begin() + k,nodeVec.end()));
	}

//
//	int k = pathwidth+1;
//	do
//	{
//      for(int pos=0; pos < length; ++pos){
//		   CNF cnf;
//		   cnf.reserve(k);
//		   for(int i=0; i < k; ++i){
//				 cnf.add(nodes[nodeVec[i]].getX(pos),false);
//		   }
//		   for(int i=0; i < graph.n(); i++){
//			   bool found = false;
//			   for(int j=0; j < k; ++j){
//				   if(nodeVec[j]==i){
//					   found =true;
//					   break;
//				   }
//			   }
//			   if(found){continue;}
//			   CNF * cnf1 = model.next(cnf.getSize()+1);  // constraint 4
//			   cnf1->add(cnf);
//			   cnf1->add(nodes[i].getX(pos),false);
//				counter++;
//		   }
//	   }
//	}
//	while(next_combinationNew(nodeVec.begin(),nodeVec.begin() + k,nodeVec.end()));
	clock.stop();

	return pwModel;
}

bool isContainedIn(ClosedInterval & i1, ClosedInterval & i2){
	return i2.first <= i1.first && i1.second <= i2.second;

}

void PWModel::checkForFeasibility(Graph& graph) {
	print(graph);

	ClosedInterval * nodeRanges = new ClosedInterval[graph.n()];

	for(int i=0; i < graph.n(); ++i){
		nodeRanges[i] = getInterval(nodes[i].getX());
	}

	ClosedInterval * edgeRanges = new ClosedInterval[graph.m()];

	for(int i=0; i < graph.m(); ++i){
		edgeRanges[i] = getInterval(edges[i]);
	}

	for(int i=0; i< graph.m(); ++i){
		Edge edge = graph.edges()[i];
		if(!isContainedIn(edgeRanges[i], nodeRanges[edge.first])||
				!isContainedIn(edgeRanges[i], nodeRanges[edge.second])){
			cout << i <<"=(" << edge.first << ","<<edge.second << ")"<<endl;
			throw "Edge is not correctly set.";
		}
	}

	int * histogram = new int[length]();

	for(int i=0; i< graph.n(); ++i){
		for(int pos=0; pos < length; ++pos){
			if(getCNFModel().getValue(nodes[i].getX(pos))){
				histogram[pos]++;
			}
		}
	}

	int maxVal =0;
	for(int pos =0; pos < length; ++pos){
		maxVal = max(maxVal,histogram[pos]);
	}
	if(maxVal != pathwidth+1){
		cout << maxVal << " " << pathwidth << endl;
		throw "Edges do not correspond with pathwidth.";
	}

}



void PWModel::print(Graph & graph){
	CNFModel & model = getCNFModel();

	cout << "nodes: " << endl;
	for(int i=0; i < graph.n(); ++i){
				cout << i << "\t";
				for(int j=0; j < length; ++j){
					if(model.getValue(nodes[i].getX(j))){
						cout << "1 ";
					}else{
						cout << "0 ";
					}
				}
				cout << endl;
			}
	cout << "edges"<<endl;
	for(int i=0; i < graph.m(); ++i){
				cout << i << "\t";
				for(int j=0; j < length; ++j){
					if(model.getValue(edges[i].getVar(j))){
						cout << "1 ";
					}else{
						cout << "0 ";
					}
				}
				cout << endl;
			}
}


