/******************************************************************************
 * STOrientation.cpp
 *
 * GDSAT
 *
 ******************************************************************************
 * Copyright (C) 2013 Benjamin Niedermann <niedermann@kit.edu>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program.  If not, see <http://www.gnu.org/licenses/>.
 *****************************************************************************/


#include "STOrientation.h"
#include "../common/Model.h"
#include "../common/SATSolver.h"
#include "../common/Box.h"





STOModel * createSTOModel(int s, int t, int height, Graph & graph, Clock& clock){
	clock.start();
	graph.addEdge(s,t);
	int length = height+2;
	int *sizes = new int[1];
	sizes[0] = height+2;

	// upper bound for variables and cnfs
	int varCount = Box::varCount(1,sizes)*graph.m()+graph.n()*length;
	int cnfCount = Box::cnfCount(1,sizes)*graph.m()
			        + graph.n()  // constraint 1
			        + graph.n()*(length-1)*length/2  // constraint 2
			        + graph.m()*length+graph.m()   // constraint 3
			        + graph.m()*length*2 // constraint 4+5

			        +graph.n()*length*2 // constraint 6+7
			        ;

	STOModel * stoModel = new STOModel(s,t,height,varCount,cnfCount);

	CNFModel & model = stoModel->getCNFModel();



	Range* nodes =  new Range[graph.n()];
	for(int i=0; i < graph.n(); ++i){
		nodes[i]=model.reserveVariables(length);
	}

	Box* edges =  new Box[graph.m()];
	for(int i=0; i < graph.m(); ++i){
		edges[i].init(model,sizes,1);
	}

	stoModel->nodes = nodes;
	stoModel->edges = edges;


	// CONSTRAINT (24): At most one grid point per node.
	for(int i=0; i < graph.n(); ++i){
		for(int k=0; k < length; ++k){
			for(int j=k+1; j < length; ++j){
				Clause * cnf1 = model.next(2);	// constraint2
				cnf1->add(nodes[i].getVar(k),false);
				cnf1->add(nodes[i].getVar(j),false);
			}
		}
	}

	// CONSTRAINT (25): At least two grid points per edge
	for(int i=0; i < graph.m(); ++i){
		Clause * cnf1 = model.next(length);
		for(int k=0; k < length; ++k){
			cnf1->add(edges[i].getX(k),true);
			Clause * cnf = model.next(length);  // constraint3
			cnf->add(edges[i].getX(k),false);

			for(int j=0; j < length; ++j){
				if(k != j){
					cnf->add(edges[i].getX(j),true);
				}
			}
		}
	}

	// CONSTRAINT (26): Edge ends and begins at node.
	for(int i=0; i< graph.m(); ++i){
		int u = graph.edges()[i].first;
		int v = graph.edges()[i].second;



		for(int j=0; j < length; ++j){
			Clause * cnf1 = model.next(3);		// constraint4
			cnf1->add(edges[i].getB(j,0),false);
			cnf1->add(nodes[u].getVar(j),true);
			cnf1->add(nodes[v].getVar(j),true);

			Clause * cnf2 = model.next(3);  // constraint5
			cnf2->add(edges[i].getE(j,0),false);
			cnf2->add(nodes[u].getVar(j),true);
			cnf2->add(nodes[v].getVar(j),true);
		}
	}

	// CONSTRAINT (27), CONSTRAINT (28): s is source, t is sink, and all other nodes are intermediate nodes.
	for(int i=0; i < graph.n(); ++i){
		if(i != t){

			for(int k=0; k< length; ++k){

			  int count =0;
			  for(int j=0; j < graph.m(); ++j){
				  count += graph.edges()[j].first == i||graph.edges()[j].second == i ? 1 :0;
			  }
			  Clause * cnf = model.next(1+count);	// constraint6
			  cnf->add(nodes[i].getVar(k),false);
			  for(int j=0; j < graph.m(); ++j){
				 if(graph.edges()[j].first == i||graph.edges()[j].second == i){

					 cnf->add(edges[j].getB(k,0),true);
				 }
			  }

			}
		}

		if(i!= s){

			for(int k=0; k < length; ++k){
				int count =0;
				for(int j=0; j < graph.m(); ++j){
					  count += graph.edges()[j].first == i||graph.edges()[j].second == i ? 1 :0;
				}
				Clause * cnf = model.next(1+count);  // constraint7
				cnf->add(nodes[i].getVar(k),false);
				for(int j=0; j < graph.m(); ++j){
					 if(graph.edges()[j].first == i||graph.edges()[j].second == i){
						 cnf->add(edges[j].getE(k,0),true);
					 }
				}

			}
		}

	}
	clock.stop();

	return stoModel;
}



int STOModel::getNodePosition(int id){
	int pos = -1;
	for(int i=0; i < length; ++i){
		if(getCNFModel().getValue(nodes[id].getVar(i))){
			if(pos != -1){
				throw "Node occupies more than one grid point";
			}
			pos = i;
		}
	}
	if(pos == -1){
		throw "Node has not been set";
	}
	return pos;
}



inline bool isEndPoint(EdgeInterval &interval, int pos){
	return interval.first == pos || pos == interval.second;
}



void STOModel::checkForFeasibility(Graph & graph){
	print(graph);

	int * incomingEdges = new int[graph.n()];
	int * outgoingEdges = new int[graph.n()];
	for(int i=0; i < graph.n(); ++i){
		incomingEdges[i] = outgoingEdges[i] =0;
	}

	for(int i=0; i < graph.m(); ++i){
		Edge edge = graph.edges()[i];

		int n1 = edge.first;
		int n2 = edge.second;

		int pos1 = getNodePosition(n1);
		int pos2 = getNodePosition(n2);

		ClosedInterval interval = getInterval(edges[i].getX());

		if(interval.second-interval.first < 1){
			throw "interval must have size of at least 2";
		}

		if(pos1 == pos2){
			cout << "edge: "<< i << "=(" << n1 << "," << n2 << ")"<< endl;
			throw "adjacent nodes may not occupy the same grid point.";
		}

		if(pos1 < pos2){
			outgoingEdges[n1]++;
			incomingEdges[n2]++;
		}else{
			outgoingEdges[n2]++;
			incomingEdges[n1]++;
		}

		if(!isEndPoint(interval,pos1) || !isEndPoint(interval,pos2)){
			cout << "edge: "<< i << "=(" << n1 << "," << n2 << ")"<< endl;
			cout << "pos: " << interval.first << " " << interval.second << endl;
			delete [] incomingEdges;
			delete [] outgoingEdges;
			throw "edge does not contain nodes correctly.";
		}
	}

	for(int i=0; i < graph.n(); ++i){
		if(i == s){
			if(outgoingEdges[i] == 0 || incomingEdges[i]>0){
				throw "s is not a source";
			}
		}
		if(i == t){
			if(incomingEdges[i] == 0 || outgoingEdges[i]>0){
				throw "t is not a sink";
			}
		}
		if(i != s && i != t){
			if(incomingEdges[i] == 0 || outgoingEdges[i]==0){
				throw "node is not a intermediate node";
			}
		}
	}
	delete [] incomingEdges;
	delete [] outgoingEdges;
}



void STOModel::print(Graph & graph){
	CNFModel & model = getCNFModel();

	cout << "nodes: " << endl;
	for(int i=0; i < graph.n(); ++i){
				cout << i << "\t";
				for(int j=0; j < length; ++j){
					if(model.getValue(nodes[i].getVar(j))){
						cout << "1 ";
					}else{
						cout << "0 ";
					}
				}
				cout << endl;
			}
	cout << "edges"<<endl;
	for(int i=0; i < graph.m(); ++i){
				cout << i << "\t";
				for(int j=0; j < length; ++j){
					if(model.getValue(edges[i].getX(j))){
						cout << "1 ";
					}else{
						cout << "0 ";
					}
				}
				cout << endl;
			}
}
