/******************************************************************************
 * tests.cpp
 *
 * GDSAT
 *
 ******************************************************************************
 * Copyright (C) 2013 Benjamin Niedermann <niedermann@kit.edu>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program.  If not, see <http://www.gnu.org/licenses/>.
 *****************************************************************************/


#include <sys/time.h>
#include <sstream>
#include <iostream>
#include <algorithm>
#include <list>
#include <fstream>
#include <iomanip>
#include "../common/Graph.h"
#include "../common/Clock.h"
#include "../common/SATSolver.h"
#include "../sat1d/Bandwidth.h"
#include "../sat1d/STOrientation.h"
#include "../sat1d/PathWidth.h"
#include "../sat2d/VisibilityRepresentation.h"

#include "../sat2d/Boxicity.h"


#include "../common/ProblemSolver.h"
#include "utils.h"



void test(Graph & graph,ProblemSolver &solver, Config * config, int value){
	Result solution;

	solver.execute(graph,solution,config);

	if(solution.model != 0){
		solution.model->checkForFeasibility(graph);
		if(value != solution.getValue()){
			cout << "expected: " << value << " computed: " <<solution.getValue() << endl;
			throw "Wrong value when checked for pathwidth";
		}
	}else{
		throw "solution could not be computed.";
	}
}

void test(string name, Graph & graph,ProblemSolver &solver,Config * config, int value){
	try{

		test(graph,solver,config,value);
	}catch(const char * & error){
		name +=" ";
		name += error;
		throw name;
	}
}

void testBC(string  name, string & graphDesc, int value){
	Graph graph = Graph();
	createGraph(graph,graphDesc);
	BCSolver solver = BCSolver();
	Config config = Config();
	test(name,graph,solver,&config,value);
}

void testPW(string  name, string & graphDesc, int value){
	Graph graph = Graph();
	createGraph(graph,graphDesc);
	PWSolver solver = PWSolver();
	Config config = Config();
	test(name,graph,solver,&config,value);
}

void testBW(string  name, string & graphDesc, int value){
	Graph graph = Graph();
	createGraph(graph,graphDesc);
	BandwidthSolver solver = BandwidthSolver();
	Config config = Config();
	test(name,graph,solver,&config,value);
}

void testVR(string  name, string & graphDesc, int value, bool oneBar){
	Graph graph = Graph();
	createGraph(graph,graphDesc);
	VRSolver solver = VRSolver();
	VRConfig config = VRConfig(oneBar);
	test(name,graph,solver,&config,value);
}

void testSTO(string  name, string & graphDesc, int value,int s, int t){
	Graph graph = Graph();
	createGraph(graph,graphDesc);
	normalize(graph,s,t);
	STOrientationSolver solver = STOrientationSolver();
	STOConfig config = STOConfig(s,t);

	test(name,graph,solver,&config,value);
}

struct RA{
	int left;
	int top;
	int right;
	int bottom;
	RA(int left, int top, int right, int bottom) : left(left),top(top), right(right), bottom(bottom){

	}

	bool contains(int row, int col){
		return top <= row && row <= bottom && left <= col && col <= right;
	}

};

void testBox(){
	const int raCount = 6;
	RA ras [] = {RA(1,3,5,6),RA(4,1,8,5),RA(2,2,4,4),RA(1,5,6,8),RA(1,1,6,1),RA(1,1,1,6)};
	Box boxes[raCount];

	int sizes [] ={10,10};

	int varCount = Box::varCount(2,sizes)*raCount;
	int cnfCount = Box::cnfCount(2,sizes)*raCount+raCount*4;

	CNFModel model = CNFModel(varCount,cnfCount);


	for(int i=0; i < raCount; ++i){
		boxes[i].init(model,sizes,2);


		model.next(1)->add(boxes[i].beginB(0)+ras[i].left,true);
		model.next(1)->add(boxes[i].beginE(0)+ras[i].right,true);
		model.next(1)->add(boxes[i].beginB(1)+ras[i].top,true);
		model.next(1)->add(boxes[i].beginE(1)+ras[i].bottom,true);
	}

	SATSolver solver;
	solver.execute(model);

	for(int i=0; i < raCount; ++i){
		cout << "Box " << i << endl;
		cout << boxes[i].toString(model) << endl;
	}

	for(int i=0; i < raCount; ++i){
		for(int row=0; row < sizes[1]; ++row){
			for(int col =0; col < sizes[0]; ++col){
				bool val =  model.getValue(boxes[i].get(row,col));
				if((val && !ras[i].contains(row,col)) || (!val && ras[i].contains(row,col))){
					throw "Box is not modeled correctly.";
				}
			}
		}
	}


}


void runTests(){
	string haus_vom_nikolaus = "5 0 1 0 2 1 3 1 2 0 3 2 3 2 4 3 4"; // 210s
	string triangle           = "3 0 1 1 2 2 0"; // < 1s
	string quadrilateral           = "4 0 1 1 2 2 3 3 0";
	string rome1             = "10 5 9 7 3 3 6 2 7 8 1 6 2 0 8 6 4 8 2 9 4"; // 517s
	string oktaeder          = "6 0 1 1 3 3 2 2 0 0 4 1 4 3 4 2 4 0 5 1 5 2 5 3 5"; //715s
	string rome2             = "20 15 0 6 10 17 3 18 13 11 4 10 8 13 19 1 4 2 13 19 7 6 18 9 1 12 13 1 18 4 16 6 16 8 5 7 14 5 15 3 18 17 1";
	string rome3 			 = "30 20 23 5 2 11 3 22 11 28 11 8 23 2 0 14 7 28 5 17 6 10 27 26 9 18 16 23 19 19 7 14 9 5 29 24 2 18 5 17 26 21 6 27 15 11 1 0 21 4 12 23 10 14 28 4 0 16 17 25 8 13 11";
	string rome4             = "24 9 18 7 10 0 22 2 3 22 1 22 18 4 20 11 18 8 1 14 15 21 17 8 7 16 19 12 6 17 2 19 7 8 5 16 2 10 17 14 18 13 23 3 12 6 7 9 13 12 0 14 1 2 4 14 13";
	string rome5			 = "22 11 17 15 20 19 12 13 10 0 13 10 2 19 0 3 0 4 16 7 18 5 20 14 9 14 6 9 8 1 19 8 21 2 17 16 2 9 5 5 1 6 18 5 8 14 18";
	string rome6             = "24 10 22 11 7 11 0 12 6 2 21 23 7 18 1 9 2 12 2 13 15 8 19 15 20 7 12 12 3 9 13 18 10 16 21 11 22 16 18 14 0 14 19 3 17 22 8 14 13 5 20 20 4 23 5 14 7";
	string rome7             = "40 20 21 27 6 24 7 27 9 15 19 27 14 25 13 33 9 12 10 7 15 18 0 35 21 10 34 36 8 17 0 4 5 11 0 23 11 27 32 39 4 33 36 13 34 28 30 31 9 27 22 6 9 37 20 26 39 32 11 0 8 25 2 38 20 11 28 30 24 0 16 18 19 16 10 25 38 33 26 8 1 5 18 27 3 27 34 29 27 15 25 9 2 31 25 27 19 20 11";
	string rome8             = "77 65 27 16 25 76 40 57 48 74 0 54 8 73 69 65 43 50 2 20 44 30 39 32 37 48 4 76 37 20 24 57 23 50 18 23 5 67 0 57 76 1 25 57 11 34 4 34 58 32 25 50 21 0 18 45 7 57 36 61 47 69 63 76 49 20 22 23 3 33 30 55 27 67 17 70 42 59 43 34 8 0 51 52 75 41 73 75 16 1 13 33 26 67 47 29 55 59 32 0 66 76 17 31 26 2 15 33 25 51 3 64 43 62 0 32 46 68 28 12 10 29 44 11 76 45 8 58 9 56 24 76 13 41 60 70 53 70 39 72 52 23 6 2 42 30 28 74 44 67 16 75 27 70 45 38 39 59 13 66 47 12 36 1 9 52 15 54 27 56 11 74 23 5 34 4 51 57 19 57 14 65 40 67 14 2 18 31 15 12 71 1 35 38 36 75 28 41 22 21 19 23 4";
	string rome9			 = "10 3 8 2 7 0 5 9 3 1 6 4 9 1 9 6 2 8 2 7 0";
	string path              = "5 0 1 1 2 2 3 3 4";
	string path2              = "8 0 1 1 2 2 3 3 4 4 5 5 6 6 7";


	string rome10            = "40 17 8 25 26 18 20 29 10 17 20 23 15 19 20 37 32 5 9 39 2 11 19 32 30 35 4 7 8 15 16 2 5 0 8 35 28 15 17 21 17 1 4 33 13 18 6 31 17 4 3 12 15 24 25 6 36 20 21 39 32 13 15 33 21 39 11 33 12 6 14 12 14 6 38 33 9 21 22 36 3 10 13 26 28 38 10 26 27 33 23 23 24 4 8 27 34 31 22 10 15";
	string rome11			 = "30 22 12 27 1 25 9 14 5 27 17 25 2 15 4 11 28 29 4 3 24 0 15 18 26 10 23 23 1 6 17 12 2 2 7 2 10 21 6 21 11 21 22 17 7 26 6 2 14 16 6 21 13 21 4 27 24 22 9 26 15 8 20 19 8 23 9 27 2 0 19 1 10 8 18 29 9 20 9 29 21";
	string rome12            = "20 16 10 17 9 1 8 17 16 2 8 8 0 5 12 17 3 14 7 9 1 6 14 12 0 10 4 17 0 0 16 17 7 17 18 17 15 0 15 5 11 17 5 15 6 19 5 6 13 18 11";
	string rome13   		 = "100 52 19 85 15 35 78 65 16 8 59 40 45 9 1 47 67 25 35 66 8 48 50 74 33 30 35 81 24 62 43 18 23 15 20 4 9 16 80 29 35 15 19 56 33 24 90 79 45 83 63 98 32 33 39 94 64 81 72 87 26 56 43 27 28 54 41 0 1 56 45 3 12 83 27 32 38 73 5 50 53 61 32 67 48 31 37 59 5 38 8 82 51 71 86 68 46 81 8 60 76 60 59 35 44 35 46 76 3 81 96 24 74 89 20 8 71 0 5 24 30 97 67 11 59 50 9 66 38 83 75 73 98 95 88 93 87 5 6 60 34 35 45 63 24 18 3 35 94 62 29 33 70 22 27 69 55 77 50 20 16 88 54 83 24 24 32 87 39 7 1 95 27 81 9 58 61 79 28 28 69 68 67 72 51 16 21 66 88 6 11 75 7 24 31 5 43 13 18 59 16 56 5 87 31 2 7 65 84 36 42 55 81 48 49 58 92 91 68 75 13 35 42 81 21 87 65 93 70 66 65 91 39 56 52 89 65 14 54 65 35 92 16 89 25 96 31 67 9 84 11 99 75 86 43 36 34 70 56 2 3 10 15 89 86 6 10 30 36 52 17 82 57";
	string rome14 			 = "38 11 24 12 10 15 6 29 18 9 36 14 10 30 17 31 20 16 13 27 1 0 27 37 4 24 0 32 28 6 20 25 0 34 3 15 21 14 23 37 8 8 32 23 0 20 2 35 34 0 8 28 7 36 28 15 22 22 14 8 29 16 12 5 30 17 0 33 16 10 13 18 5 11 16 28 6 37 31 1 19 5 34 35 31 26 14 31 6 17 4 35 10 25 21 31 19";

	string pathwidth2 		 = "7 0 1 1 2 0 2 0 3 3 2 2 5 2 4 3 4 3 5 5 6";
	string pathwidth3        = "4 0 1 0 2 1 2 0 3 1 3 2 3";
	string pathwidthB3        = "8 0 1 0 2 0 3 1 4 2 5 3 6 4 7 5 7 6 7";
	string pathwidthC3	     = "6 0 1 1 2 0 3 1 3 1 4 2 4 3 4 3 5 4 5";
	string pathwidth4	     = "8 0 1 1 2 2 3 3 0 0 4 1 5 2 6 3 7 4 5 5 6 6 7 7 4";
	string pathwidthB4       = "7 0 1 1 2 2 3 3 4 4 5 5 0 5 6 1 6 3 6 0 2 2 4 4 0";
	string pathwidthC4		 = "9 0 1 1 2 2 3 3 4 4 5 5 0 0 6 6 3 1 8 8 4 2 7 7 5";
	string pathwidthD4       = "9 0 1 0 2 0 3 1 4 2 5 3 6 4 7 5 7 6 7 1 8 2 8 5 8 6 8";


	string graph1 			 = "4 0 1 0 2 2 1 2 3 3 1 0 3";
	string prob              = "11 8 3 10 3 6 0 5 2 3 7 1 4 10 7 9 5 1 9 7 2 10 1 6 5 5 3";

	string rome15			 = "10 1 9 0 5 2 7 6 2 8 2 7 0 9 3 3 8 4 9 1 6";
	string rome16			 = "40 11 19 39 2 17 8 12 14 20 21 21 22 10 13 12 15 33 12 35 28 2 5 26 28 33 21 1 4 7 8 33 23 6 14 18 6 31 17 10 15 23 24 21 17 4 8 17 20 33 9 13 15 36 3 5 9 26 27 6 36 31 22 37 32 4 3 27 34 23 15 15 17 33 13 0 8 18 20 35 4 19 20 39 32 29 10 6 38 38 10 15 16 39 11 32 30 25 26 24 25";
	string rome17			 = "30 2 7 25 9 21 22 25 2 22 12 1 10 26 6 27 2 21 11 23 9 23 1 21 6 11 28 12 2 17 7 0 19 20 9 26 15 8 18 2 14 29 9 22 9 19 8 2 10 8 20 10 23 21 4 18 26 15 4 29 4 0 15 21 13 3 24 14 5 29 21 6 17 27 1 16 6 27 24 27 17";
	string rome18			 = "38 25 0 23 0 12 10 14 10 5 34 11 24 28 7 14 23 28 6 11 16 37 4 24 0 35 31 15 21 30 17 36 28 9 36 20 2 25 21 37 31 1 19 17 4 35 34 5 30 26 14 33 16 31 6 16 12 34 3 18 5 17 0 0 27 10 13 22 14 27 1 31 20 0 8 15 22 16 13 35 10 31 19 8 32 6 20 15 6 37 8 32 28 8 29 29 18";

	string k4           = "4 0 1 1 2 2 3 3 0 0 2 1 3";
	string k5           = "5 0 1 1 2 2 3 3 0 0 2 1 3 0 4 1 4 2 4 3 4";
	string graph2       = "5 0 1 0 2 1 3 1 2 0 3 2 3 2 4 3 4";



	try{
		cout << "RUN TESTS FOR PATHWIDTH" << endl;
		testPW("PATHWIDTH1",path,1);
		testPW("PATHWIDTH2",pathwidth2,2);
		testPW("PATHWIDTH3",pathwidth3,3);
		testPW("PATHWIDTH4",pathwidthB3,3);
		testPW("PATHWIDTH5",pathwidthC3,3);
		testPW("PATHWIDTH6",pathwidth4,4);
		testPW("PATHWIDTH7",pathwidthB4,4);
		testPW("PATHWIDTH8",pathwidthC4,4);
		testPW("PATHWIDTH9",pathwidthD4,4);
		testPW("PATHWIDTH10",triangle,2);
		testPW("PATHWIDTH11",quadrilateral,2);



		cout << "RUN TESTS FOR BANDWIDTH" << endl;
		testBW("BANDWIDTH1",triangle,2);
		testBW("BANDWIDTH2",quadrilateral,2);
		testBW("BANDWIDTH3",k4,3);
		testBW("BANDWIDTH4",k5,4);
		testBW("BANDWIDTH5",path,1);
		testBW("BANDWIDTH6",rome1,2);
		testBW("BANDWIDTH7",rome9,2);
		testBW("BANDWIDTH8",rome15,2);
		testBW("BANDWIDTH9",oktaeder,4);
		testBW("BANDWIDTH10",graph2,3);

		cout << "RUN TESTS FOR ST-ORIENTATIONS" << endl;
		testSTO("STO1",rome15,8,4,5);
		testSTO("STO2",rome16,15,34,36);
		testSTO("STO3",rome11,15,3,28);
		testSTO("STO4",rome12,9,4,19);
		testSTO("STO5",triangle,3,0,1);
		testSTO("STO6",quadrilateral,3,0,2);
		testSTO("STO7",k4,4,0,2);
		testSTO("STO8",k5,5,0,2);
		testSTO("STO9",oktaeder,5,0,2);

		cout << "RUN TESTS FOR BOXICITY" << endl;
		// value+2 because each grid has a frame.
		testBC("BC1",triangle,1+2);
		testBC("BC2",quadrilateral,2+2);
		testBC("BC3",k4,1+2);
		testBC("BC4",k5,1+2);
		testBC("BC5",path,2+2);
		testBC("BC6",rome1,4+2);
		testBC("BC7",rome9,4+2);
		testBC("BC8",path,2+2);
		testBC("BC9",path2,3+2);

		testVR("VR1A",triangle,3+2,false);
		testVR("VR1B",triangle,3+2,true);
		testVR("VR2A",quadrilateral,3+2,false);
		testVR("VR2B",quadrilateral,3+2,true);
		testVR("VR3A",path,3+2,false);
		testVR("VR3B",path,3+2,true);
		testVR("VR4A",path2,4+2,false);
		testVR("VR4B",path2,4+2,true);
		testVR("VR5A",graph1,4+2,false);
		testVR("VR5B",graph1,4+2,true);
		testVR("VR6A",rome1,5+2,false);
		testVR("VR6B",rome1,5+2,true);


		testBox();


	}catch(string & error){
			cout << "ERROR: " << error << endl;
	}

	cout << "Tests have been executed!" << endl;
}




